/* basicio.c
   Copyright (C) 1997 L.C. Benschop, Eindhoven, the Netherlands
   */

#include "gplbasic.h"
#include "basicio.h"
#include "errors.h"
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

typedef struct {
  FILE * fileptr;
  int mode; /* I O A R B open mode */
  int outcol; /* output column (for ,) or record length.*/
} BASICFILE;

BASICFILE basicfile[MAX_FILES];

int cur_input,cur_output; /* current input and output channels ,0 = tty,
                    1..MAX_FILES = file */

static char *inpptr;

static int output_col;

static char spaces[80];

static void outstring(unsigned char *,int);
static void termoutstring(unsigned char *,int);
static void terminstring(unsigned char *,int);

void init_io(void) 
{
  int i;
  for(i=0;i<80;i++) spaces[i]=' ';
  output_col=0;
  cur_input=cur_output=0;
}

void print_int(long l)
{
  int len;
  len=sprintf(progline,"% ld ",l);
  outstring(progline,len);
}

void print_str(unsigned char *p)
{
  int n;
  n=*p++;
  n=n+256*(*p++ & 127); /* get string length. */
  outstring(p,n);
}    

void write_str(unsigned char *p)
{
  int n,k;
  n=*p++;
  n=n+256*(*p++ & 127); /* get string length. */
  progline[0]='"';k=1;
  while(n) {
    progline[k++]=*p++;
    if(*(p-1)=='"') progline[k++]='"';
    if(k>=MAXLINELENGTH) {
      outstring(progline,k);k=0;
    }
    n--;
  }
  progline[k++]='"';
  outstring(progline,k);
}


void print_float(double f)
{
  int len;
  len=sprintf(progline,"% .10g ",f);
  outstring(progline,len);
}

void print_nl(void)
{
  outstring("\n",1);
  if(cur_output) basicfile[cur_output-1].outcol=0;
   else output_col=0;
}


void print_tab(int t)
{
  int oc;
  if(cur_output) oc=basicfile[cur_output-1].outcol;
   else oc=output_col;
  if(oc+1>t) print_nl();
  while(t-oc-1>80) {
    outstring(spaces,80);
    t-=80;
  } 
  outstring(spaces,t-oc-1);  
}


void print_com(void)
     /* Comma operator in PRINT statement */
{
  int oc;
  if(cur_output) oc=basicfile[cur_output-1].outcol;
   else oc=output_col;
  outstring(spaces,15-(oc+15)%16);
}

void input_nl(void)
{
  if(cur_input) {
    fgets(progline,MAXLINELENGTH+2,basicfile[cur_input-1].fileptr);
  } else {
    terminstring(progline,MAXLINELENGTH);
    output_col=0;
  }
  if(strlen(progline)) progline[strlen(progline)-1]=0;
  inpptr=progline;
}

long input_int(void)
{
  return atol(inpptr);
}

double input_float(void)
{
  return atof(inpptr);
}

void input_com(void)
{
  while(*inpptr!=',' && *inpptr!=0) inpptr++;
  if(*inpptr==',') inpptr++;
}

unsigned char * input_str(void)
{
  unsigned char *r;
  int rl;
  rl=strlen(inpptr);
  if(string_stack+S_STRINGSTACK-strstkptr < rl+2) doerror(GE_STS,0,0);
  r=strstkptr;
  while(*inpptr==' '||*inpptr=='\t')inpptr++;
  rl=0;
  if(*inpptr=='"') { /* input string within quotes */
    inpptr++;
    while(*inpptr && (*inpptr!='"' || inpptr[1]=='"')) {
      if(*inpptr=='"') inpptr++;
      r[2+rl++]=*inpptr++;
    }
  } else { /* no quotes around string */
    while(*inpptr && *inpptr!=',') {
      r[2+rl++]=*inpptr++;
    }
    while(rl && (r[rl+1]==' ' || r[rl+1]=='\t')) rl--; 
    /* get rid of trailing blanks */
  }
  strstkptr+=rl+2;
  *r=rl&255;
  *(r+1)=rl>>8;
  return r;
}

unsigned char * linput_str(void)
{
  unsigned char *r;
  int rl;
  rl=strlen(progline);
  if(string_stack+S_STRINGSTACK-strstkptr < rl+2) doerror(GE_STS,0,0);
  r=strstkptr;
  strstkptr+=rl+2;
  memcpy(r+2,progline,rl);
  *r=rl&255;
  *(r+1)=rl>>8;
  return r;
}

void outstring(unsigned char *p,int len)
{
  if(cur_output) {
    fwrite(p,1,len,basicfile[cur_output-1].fileptr);
    basicfile[cur_output-1].outcol+=len;
  } else {
    termoutstring(p,len);
    output_col+=len;
  }
}

void termoutstring(unsigned char *p,int len)
{
  fwrite(p,1,len,stdout);
}

void terminstring(unsigned char *p,int len)
{
  fgets(p,len+1,stdin);
}

void do_fopen(unsigned char *mode,int fn,unsigned char * name)
{
  int nl; /* name length */
  int m;
  FILE * fp;
  if(fn<1||fn>MAX_FILES) doerror(GE_BFN,0,0);
  if(basicfile[fn-1].fileptr!=0) doerror(GE_BFN,0,0);
  nl=*name+256*(name[1]&127);
  if(nl>MAXLINELENGTH) doerror(GE_BFN,0,0);
  memcpy(progline,name+2,nl);
  progline[nl]=0;
  if(*mode!=1 || ((mode[1]&127)!=0)) doerror(GE_BFN,0,0);
  switch(mode[2]) {
  case 'I':case 'i':
    fp=fopen(progline,"r");m='I';
    if(!fp)doerror(GE_OFR,progline,0);
  break;
  case 'O':case 'o':
    fp=fopen(progline,"w");m='O';
    if(!fp)doerror(GE_OFW,progline,0);
  break;
  case 'A':case 'a':
    fp=fopen(progline,"a");m='A';
    if(!fp)doerror(GE_OFA,progline,0);
  break;
  default:doerror(GE_BFN,0,0);
  }
  basicfile[fn-1].fileptr=fp;
  basicfile[fn-1].mode=m;
}

void do_fclose(int fn)
{
  if(fn<1||fn>MAX_FILES) doerror(GE_BFN,0,0);
  if(basicfile[fn-1].fileptr==0) doerror(GE_BFN,0,0);
  fclose(basicfile[fn-1].fileptr);
  basicfile[fn-1].fileptr=0;
}

void set_input(int fn) 
{ 
  if(fn==0) {
    cur_input=0;
  } else {
    if(fn<1||fn>MAX_FILES) doerror(GE_BFN,0,0);
    if(basicfile[fn-1].fileptr==0||basicfile[fn-1].mode!='I') 
      doerror(GE_BFN,0,0);
    cur_input=fn;
  }
}

void set_output(int fn)
{
  if(fn==0) {
    cur_output=0;
  } else {
    if(fn<1||fn>MAX_FILES) doerror(GE_BFN,0,0);
    if(basicfile[fn-1].fileptr==0||
       (basicfile[fn-1].mode!='O'&&basicfile[fn-1].mode!='A')) 
      doerror(GE_BFN,0,0);
    cur_output=fn;
  }
}

int do_eof(int fn)
{
  int c;
  if(fn<1||fn>MAX_FILES) doerror(GE_BFN,0,0);
  if(basicfile[fn-1].fileptr==0||basicfile[fn-1].mode!='I') 
    doerror(GE_BFN,0,0);
  c=fgetc(basicfile[fn-1].fileptr);
  if(c!=EOF) ungetc(c,basicfile[fn-1].fileptr);
  return -(c==EOF);
}

int do_lof(int fn)
{
  long pos,size;
  if(fn<1||fn>MAX_FILES) doerror(GE_BFN,0,0);
  if(basicfile[fn-1].fileptr==0||basicfile[fn-1].mode!='I') 
    doerror(GE_BFN,0,0);
  pos=ftell(basicfile[fn-1].fileptr);
  fseek(basicfile[fn-1].fileptr,0,SEEK_END);
  size=ftell(basicfile[fn-1].fileptr);
  fseek(basicfile[fn-1].fileptr,pos,SEEK_SET);
  return size;
}

int do_freefile(void)
{
  int i;
  for(i=0;i<MAX_FILES;i++) {
    if(basicfile[i].fileptr==0) return i+1;
  }
  return -1;
}
