/* gplbasic.h
   Copyright (C) 1997 L.C. Benschop, Eindhoven, the Netherlands
   */

#include <setjmp.h>

typedef union {
         long l;
         void * p;
} CELLTYPE;

/* Size configuration options */

#define N_MEMAREAS 256   /* Number of allocated memory areas. */
#define S_STRINGSTACK 65536 /* Size of string stack */
#define S_FLTSTACK 256      /* Size of floating point stack */
#define S_MAINSTACK 1024    /* Size of main stack */
#define S_STRINGAREA 65536 /* Size of strings area */
#define S_PROGCHUNK 8192 /* Size of program chunk (in cells) */
#define S_SYMCHUNK  256 /* Size of symbol table chunk (in entries) */
#define S_STRCONST 4096 /* Size of string constant area chunk */
#define S_FCONST 256   /* Size of floating constant area chunk */
#define S_HASHENTRIES 233 /* Size of (open) hash table */
#define MAX_FILES 20      /* Maximum number of open files */
#define MAXLINELENGTH 260 /* Maximum line length in source file */

extern void * mem_areas[N_MEMAREAS]; /* Allocated memory areas */
extern unsigned char string_stack[S_STRINGSTACK]; 
extern double float_stack[S_FLTSTACK];
extern CELLTYPE main_stack[S_MAINSTACK];
extern CELLTYPE * progstart;

extern CELLTYPE * progptr;
extern CELLTYPE * stkptr;
extern double   * fstkptr;
extern unsigned char * strstkptr;  
extern unsigned char * strstkmark;
extern CELLTYPE *frameptr;
extern CELLTYPE *stframe; /* frame pointer for local strings */
extern double *flframe; /* frame pointer for FP numbers */

extern int memareas; /* number of memory areas in use*/
extern int charareas; /* Number of char const areas */
extern unsigned char * charconst; /* ptr to character constant area */
extern int chars_used; /* Number of character constants used */
extern int progareas; /* Number of program memory areas */
extern int cells_used; /* Number of cells used in current program area*/ 
extern int floatareas; /* Number of floating constant areas */
extern int floats_used; /* Number of floats used in current area */
extern int symareas; /* Number of symbol table areas */
extern int syms_used; /* Number of symbols used in current area */
extern int labareas;  /* Number of label table areas */
extern int labs_used; /* Number of labels used in current area */

extern unsigned char * stringareas; /* pointer to first string area */
extern CELLTYPE * datastart; /* pointer to first DATA statement. */
extern CELLTYPE * cur_data; /* pointer to current data statement */
extern int data_used; /* number of data items used in current statement */

void clear_vars(void); /* Clear all variables (and stacks) */
int load_program(char *); /* Load (and compile) program */
int run_program(CELLTYPE *); /* Run a program */
void debugger(void); /* Program debugger */

void init_keywords(int);
void init_io(void);
void init_strings(void);

extern char progline[]; /* program source line */
extern long lineno; /* source line number */

extern char progname[]; /* program name */

extern jmp_buf errorbuf;

CELLTYPE * last_line; /* last executed program line */
int errhandling;

