/* runner.c   The GPLBASIC interpreter
   Copyright (C) 1997 L.C. Benschop, Eindhoven, the Netherlands
   */

#include "gplbasic.h"
#include "opcodes.h"
#include "token.h"
#include "errors.h"
#include "basicio.h"
#include "strings.h"
#include <stdlib.h>
#include <math.h>
#include <time.h>

CELLTYPE * progptr;
CELLTYPE * stkptr;
double   * fstkptr;
unsigned char * strstkptr;  
unsigned char * strstkmark;
int data_used;
CELLTYPE * frameptr;
CELLTYPE * stframe;
double * flframe;

CELLTYPE *progstart;

void * index_array(int);
void dim_array(int,CELLTYPE**,int);
static int bas_timer(void);

#define ipop(n) do{t=*stkptr++;n=t.l;}while(0)
#define ipush(n) do{if(stkptr==main_stack) doerror(GE_STK,0,0); else\
  {t.l=n;*--stkptr=t;}}while(0)
#define ppop(n) do{t=*stkptr++;n=t.p;}while(0)
#define ppush(n) do{if(stkptr==main_stack) doerror(GE_STK,0,0); else\
  {t.p=n;*--stkptr=t;}}while(0)
#define fpop(n) (n=*fstkptr++)
#define fpush(n) do{if(fstkptr==float_stack) doerror(GE_STK,0,0); else\
  {*--fstkptr=n;}}while(0)

int run_program(CELLTYPE *start)
{
  CELLTYPE opcode;
  CELLTYPE t;
  long a,b,c;
  CELLTYPE *p;
  CELLTYPE **ap;
  unsigned char *ca,*cb;
  double fa,fb,fc;
  if( start == 0 ) progptr=progstart; else progptr=start;
  for(;;) {
    opcode= *progptr++;
    switch( opcode.l ) {
    case OP_IADD: ipop(b); stkptr->l+=b; break;
    case OP_ISUB: ipop(b); stkptr->l-=b; break;
    case OP_IMUL: ipop(b); stkptr->l*=b;  break;
    case OP_IDIV: ipop(b); 
            if( b==0 ) doerror(GE_DIV,0,0);
            stkptr->l/=b; break;
    case OP_IMOD: ipop(b);  
            if( b==0 ) doerror(GE_DIV,0,0);
	    stkptr->l%=b;  break;
    case OP_IAND: ipop(b); stkptr->l&=b;  break;
    case OP_IOR:  ipop(b); stkptr->l|=b;  break;
    case OP_IXOR: ipop(b); ipop(a); a=a^b; ipush(a); break;
    case OP_IEQV: ipop(b); ipop(a); a=~(a^b); ipush(a); break;
    case OP_IIMP: ipop(b); ipop(a); a=(~a)|b; ipush(a); break;
    case OP_ILIT: a=progptr->l; progptr++; ipush(a); break;
    case OP_IGET: a=*(long *)(progptr->p); progptr++; ipush(a); break;
    case OP_IPUT: ipop(a); *(long *)(progptr->p)=a; progptr++; break;
    case OP_IEQ: ipop(b); ipop(a); a=-(a==b); ipush(a); break;
    case OP_INE: ipop(b); ipop(a); a=-(a!=b); ipush(a); break;
    case OP_ILT: ipop(b); ipop(a); a=-(a<b); ipush(a); break;
    case OP_IGE: ipop(b); ipop(a); a=-(a>=b); ipush(a); break;
    case OP_IGT: ipop(b); ipop(a); a=-(a>b); ipush(a); break;
    case OP_ILE: ipop(b); ipop(a); a=-(a<=b); ipush(a); break;
    case OP_IPRINT: ipop(a); print_int(a); break;
    case OP_PRINTNL: print_nl(); break;
    case OP_BT: ipop(a); if(a) progptr=progptr->p; else progptr++; break;
    case OP_BF: ipop(a); if(!a) progptr=progptr->p; else progptr++; break;
    case OP_JMP:  progptr=progptr->p; break;
    case OP_EXIT: return 0; break;
    case OP_INEG: ipop(a); a=-a; ipush(a); break;
    case OP_INOT: ipop(a); a=~a; ipush(a); break;
    case OP_JSR: p=progptr+1;ppush(p);ipush(KW_RETURN);
                 progptr=progptr->p; break;
    case OP_RET: 
      while(stkptr->l!=KW_RETURN) { /* ret rid of any unterminated fors*/
	/* This is an UNLY KLUDGE!!!! But I had to play Wumpus! */
	if(stkptr->l==KW_FOR) stkptr+=5;
	else if(stkptr->l==KW_NEXT) {stkptr+=3;fstkptr+=2;}
	else doerror(GE_WIT,"RETURN","GOSUB");
      }
      stkptr++; ppop(progptr); 
    break;
    case OP_PRINTC: print_com(); break;
    case OP_SPRINT: ppop(ca);print_str(ca);break;
    case OP_SLIT: ca=progptr->p;progptr++;ppush(ca);break;
    case OP_SGET: ca=*(unsigned char **)(progptr->p);progptr++;ppush(ca);
                  break;
    case OP_SPUT: ppop(ca); put_str((unsigned char**)(progptr->p),ca);
                  progptr++;break;
    case OP_CHR: ipop(a);ca=short_strings+2+3*a;ppush(ca);break;
    case OP_LEN: ppop(ca);a=*ca+256*(ca[1]&127);ipush(a);break;
    case OP_ASC: ppop(ca);a=*ca+256*(ca[1]&127);if(a)a=ca[2];
                 ipush(a);break;
    case OP_IABS: ipop(a);a=abs(a);ipush(a);break;
    case OP_ISGN: ipop(a);if(a<0)a=-1; else if(a>0) a=1; ipush(a);break;
    case OP_INPUTNL: input_nl();break;
    case OP_IINPUT: a=input_int();ipush(a);break;
    case OP_IFOR:
      p=progptr->p;
    lab_ifor:
      ipop(c);ipop(b);ipop(a);
      *(long *)p=a;
      if((a>b && c>0) || (c<0 && a<b)) { /* go through loop zero times */
       progptr=progptr[1].p;
      } else {
       ppush(p);
       progptr+=2;
       ipush(b);
       ipush(c);
       ppush(progptr);
       ipush(KW_FOR);
      }
    break;
    case OP_INEXT:
      if(stkptr->l !=KW_FOR) doerror(GE_WIT,"NEXT","FOR");
      c=stkptr[2].l; /*stepsize*/
      b=stkptr[3].l; /*limit*/
      a=*(long *)stkptr[4].p+=c; /* add step to variable */
      if((a>b && c>=0) || (c<0 && a<b)) { /* terminate loop */
      	stkptr+=5;
      } else {
        progptr=stkptr[1].p;
      }
    break;
    case OP_TJMP:
      ipop(b);a=progptr->l;if(b<=0 ||b>a) progptr+=a+1; else 
	progptr=progptr[b].p;
    break;
    case OP_TJSR:
      ipop(b);a=progptr->l;if(b<=0 ||b>a) progptr+=a+1; else {
	p=progptr+a+1;ppush(p);progptr=progptr[b].p;
        ipush(KW_RETURN);
      }
    break;
    case OP_IGETA:
      a=*(long*)index_array(TY_INT);
      ipush(a);
      progptr+=2;
    break;
    case OP_IPUTA:
      ipop(a);
      *(long*)index_array(TY_INT)=a;
      progptr+=2;
    break;
    case OP_IDIM:
      a=progptr->l; /* number of dimensions */
      ap=progptr[1].p; /* pointer to variable */
      if(*ap) doerror(GE_DIM,0,0);
      dim_array(a,ap,TY_INT);
      progptr+=2;
    break;
    case OP_FADD: fpop(fa);*fstkptr+=fa;break;
    case OP_FSUB: fpop(fa);*fstkptr-=fa;break;
    case OP_FMUL: fpop(fa);*fstkptr*=fa;break;
    case OP_FDIV: fpop(fa);*fstkptr/=fa;break;
    case OP_FGET: fa=*(double *)(progptr->p);progptr++;fpush(fa);break;
    case OP_FPUT: fpop(fa);*(double *)(progptr->p)=fa;progptr++;break;
    case OP_FEQ: fpop(fb);fpop(fa);a=-(fa==fb);ipush(a);break;
    case OP_FNE: fpop(fb);fpop(fa);a=-(fa!=fb);ipush(a);break;
    case OP_FLT: fpop(fb);fpop(fa);a=-(fa<fb);ipush(a);break;
    case OP_FGE: fpop(fb);fpop(fa);a=-(fa>=fb);ipush(a);break;
    case OP_FGT: fpop(fb);fpop(fa);a=-(fa>fb);ipush(a);break;
    case OP_FLE: fpop(fb);fpop(fa);a=-(fa<=fb);ipush(a);break;
    case OP_FIPOW:                  break;
    case OP_FFPOW: fpop(fb);fpop(fa);fa=pow(fa,fb);fpush(fa); break;
    case OP_I2F: ipop(a);fa=a;fpush(fa);break;
    case OP_F2I: fpop(fa);a=fa;ipush(a);break;
    case OP_I2F2: fpop(fb);ipop(a);fa=a;fpush(fa);fpush(fb);break;
    case OP_FPRINT: fpop(fa);print_float(fa);break;
    case OP_FGETA:
      fa=*(double*)index_array(TY_DOUBLE);
      fpush(fa);
      progptr+=2;
    break;
    case OP_FDIM:
      a=progptr->l; /* number of dimensions */
      ap=progptr[1].p; /* pointer to variable */
      if(*ap) doerror(GE_DIM,0,0);
      dim_array(a,ap,TY_DOUBLE);
      progptr+=2;
    break;
    case OP_FPUTA:
      fpop(fa);
      *(double*)index_array(TY_DOUBLE)=fa;
      progptr+=2;
    break;
    case OP_FINPUT: fa=input_float();fpush(fa);break;
    case OP_FFOR:
      p=(progptr->p);
    lab_ffor:
      fpop(fc);fpop(fb);fpop(fa);
      *(double *)p=fa;
      if((fa>fb && fc>0) || (fc<0 && fa<fb)) { /* go through loop zero times */
       progptr=progptr[1].p;
      } else {
       ppush(p);
       progptr+=2;
       fpush(fb);
       fpush(fc);
       ppush(progptr);
       ipush(KW_NEXT);
      }
    break;
    case OP_FNEXT:
      if(stkptr->l !=KW_NEXT) doerror(GE_WIT,"NEXT","FOR");
      fc=fstkptr[0]; /*stepsize*/
      fb=fstkptr[1]; /*limit*/
      fa=*(double *)stkptr[2].p+=fc; /* add step to variable */
      if( (fa>fb && fc>=0) || (fc<0 && fa<fb)) { /* terminate loop */
      	stkptr+=3;
	fstkptr+=2;
      } else {
        progptr=stkptr[1].p;
      }
    break;
    case OP_SIN: fpop(fa);fa=sin(fa);fpush(fa); break;
    case OP_COS: fpop(fa);fa=cos(fa);fpush(fa); break;
    case OP_TAN: fpop(fa);fa=tan(fa);fpush(fa); break;
    case OP_ATN: fpop(fa);fa=atan(fa);fpush(fa); break;
    case OP_SQR: fpop(fa);fa=sqrt(fa);fpush(fa); break;
    case OP_LOG: fpop(fa);fa=log(fa);fpush(fa); break;
    case OP_EXP: fpop(fa);fa=exp(fa);fpush(fa); break;
    case OP_FNEG: *fstkptr=-*fstkptr;break;
    case OP_FABS: *fstkptr=fabs(*fstkptr);break;
    case OP_FSGN: fpop(fa); if(fa<0)a=-1; else if(fa>0) a=1; else a=0;
                  ipush(a);break;
    case OP_INT: fpop(fa);fa=floor(fa);fpush(fa); break;
    case OP_RND: fpop(fa);fa=rand()/((double)RAND_MAX+1);fpush(fa);break;
    case OP_CLR: clear_vars();break;
    case OP_SEQ: ppop(cb);ppop(ca);a=-(comp_str(ca,cb)==0);ipush(a);break;
    case OP_SNE: ppop(cb);ppop(ca);a=-(comp_str(ca,cb)!=0);ipush(a);break;
    case OP_SLT: ppop(cb);ppop(ca);a=-(comp_str(ca,cb)<0);ipush(a);break;
    case OP_SGE: ppop(cb);ppop(ca);a=-(comp_str(ca,cb)>=0);ipush(a);break;
    case OP_SGT: ppop(cb);ppop(ca);a=-(comp_str(ca,cb)>0);ipush(a);break;
    case OP_SLE: ppop(cb);ppop(ca);a=-(comp_str(ca,cb)<=0);ipush(a);break;
    case OP_CLRSTK: strstkptr=strstkmark;break;
    case OP_SCAT: ppop(cb);ppop(ca);ca=cat_str(ca,cb);ppush(ca);break;
    case OP_LEFT: ipop(a);ppop(ca);ca=left_str(ca,a);ppush(ca);break;
    case OP_MID: ipop(b);ipop(a);ppop(ca);ca=mid_str(ca,a,b);ppush(ca);break;
    case OP_MID2: ipop(a);ppop(ca);ca=mid_str2(ca,a);ppush(ca);break;
    case OP_RIGHT: ipop(a);ppop(ca);ca=right_str(ca,a);ppush(ca);break;
    case OP_STR: fpop(fa);ca=str_str(fa);ppush(ca);break;
    case OP_VAL: ppop(ca);fa=val_str(ca);fpush(fa);break;
    case OP_STRING: ipop(b);ipop(a);ca=string_str(a,b);ppush(ca);break;
    case OP_INSTR: ppop(cb);ppop(ca);ipop(a);a=instr(a,ca,cb);ipush(a);break;
    case OP_SINPUT: ca=input_str();ppush(ca);break;
    case OP_SGETA:
      ca=*(unsigned char **)index_array(TY_STR);
      ppush(ca);
      progptr+=2;
    break;
    case OP_SPUTA:
      ppop(ca);
      put_str((unsigned char**)index_array(TY_STR),ca);
      progptr+=2;
    break;
    case OP_SDIM:
      a=progptr->l;
      ap=progptr[1].p;
      if (*ap) doerror(GE_DIM,0,0);
      dim_array(a,ap,TY_STR);
      progptr+=2;
    break;
    case OP_SRAND: ipop(a); if(a==-1) a=time(0); srand(a); break;
    case OP_RESTOR: p=progptr->p;progptr++; 
      if(!p) {
	cur_data=datastart;
      } else {
	cur_data=p;
      }
      for(;;) {
	if(cur_data->l==OP_JMP) cur_data=cur_data[1].p;
        else if(cur_data->l==OP_MARKLINE) cur_data=cur_data+4;
        else if(cur_data->l==OP_SKDATA) break;
	else doerror(GE_DAT,0,0);
      }
      data_used=0;
    break;
    case OP_READ:
      if(cur_data==0) doerror(GE_DAT,0,0);
      if ((CELLTYPE*)(cur_data[1].p)-cur_data == 3+data_used) { 
	/* all data in current statement is used */
	cur_data=cur_data[2].p;
	if(cur_data==0) doerror(GE_DAT,0,0);
	data_used=0;
      }
      p=cur_data[3+data_used++].p;
      ppush(p);
    break;
    case OP_PRINTT: ipop(a); print_tab(a); break;
      /* local variable access */
    case OP_IGETL: a=progptr->l;progptr++;a=(frameptr+a)->l;ipush(a);break;
    case OP_IPUTL: ipop(b);a=progptr->l;progptr++;(frameptr+a)->l=b;break;
    case OP_FGETL: a=progptr->l;progptr++;fa=*(flframe+a);fpush(fa);break;
    case OP_FPUTL: fpop(fa);a=progptr->l;progptr++;*(flframe+a)=fa;break;
    case OP_SGETL: a=progptr->l;progptr++;ca=(stframe+a)->p;ppush(ca);break;
    case OP_SPUTL: ppop(ca);a=progptr->l;progptr++;
     put_str((unsigned char **)(stframe+a),ca);break; 
      /* reference parameter access */
    case OP_IGETR: a=progptr->l;progptr++;a=*(long *)((frameptr+a)->p);
      ipush(a);break;
    case OP_IPUTR: ipop(b);a=progptr->l;progptr++;
      *(long *)((frameptr+a)->p)=b;break;
    case OP_FGETR: a=progptr->l;progptr++;fa=*(double *)((frameptr+a)->p);
      fpush(fa);break;
    case OP_FPUTR: fpop(fa);a=progptr->l;progptr++;
      *(double *)((frameptr+a)->p)=fa;break;
    case OP_SGETR: a=progptr->l;progptr++;ca=*(char **)((frameptr+a)->p);
      ppush(ca);break;
    case OP_SPUTR: ppop(ca);a=progptr->l;progptr++;
      put_str((unsigned char **)((frameptr+a)->p),ca);break;
      /* subroutine/function call */
    case OP_CALL: p=progptr+1;ppush(p);progptr=progptr->p;break;
    case OP_LINK: /* create new stack frame */
      ppush(frameptr);ppush(stframe);
      ppush(flframe);ppush(strstkmark);frameptr=stkptr;
      flframe=fstkptr;
      for(a=0;a<progptr->l;a++)ipush(0);
      stframe=stkptr;
      for(a=0;a<progptr[1].l;a++)ppush(short_strings);
      for(a=0;a<progptr[2].l;a++)fpush(0.0);
      strstkmark=strstkptr;
      progptr+=3;
      ipush(KW_CALL);
      break;
    case OP_UNLRET:
      while(stkptr->l!=KW_CALL) { /* ret rid of any unterminated fors*/
	/* This is an UNLY KLUDGE */
	if(stkptr->l==KW_FOR) stkptr+=5;
	else if(stkptr->l==KW_NEXT) {stkptr+=3;fstkptr+=2;}
	else doerror(GE_WIT,"END SUB","CALL");
      }
      stkptr++;
      while(stkptr!=stframe) {
	put_str((unsigned char**)stkptr ,short_strings); 
	/* unbind local variables out of heap */
	stkptr++;
      }
      fstkptr=flframe;
      stkptr=frameptr;
      ppop(strstkmark);ppop(flframe);ppop(stframe);ppop(frameptr);ppop(p);
      stkptr+=progptr->l;
      fstkptr+=progptr[1].l; /* remove parameters from stack */
      progptr=p;break;
    case OP_SWAP: t=stkptr[0];stkptr[0]=stkptr[1];stkptr[1]=t;break;
    case OP_FZERO: fpush(0);break;
    case OP_FDROP: fstkptr++;break;
    case OP_TIME: ca=time_str(); ppush(ca); break; 
    case OP_DATE: ca=date_str(); ppush(ca); break;
    case OP_TIMER: a=bas_timer(); ipush(a); break;
    case OP_OCT: ipop(a);ca=oct_str(a);ppush(ca);break;
    case OP_HEX: ipop(a);ca=hex_str(a);ppush(ca);break;
    case OP_UCASE: ppop(ca);ca=ucase_str(ca);ppush(ca);break;
    case OP_LCASE: ppop(ca);ca=lcase_str(ca);ppush(ca);break;
    case OP_LTRIM: ppop(ca);ca=ltrim_str(ca);ppush(ca);break;
    case OP_RTRIM: ppop(ca);ca=rtrim_str(ca);ppush(ca);break;
    case OP_EXITFOR: 
      if(stkptr->l==KW_FOR) {
	progptr=((CELLTYPE*)stkptr[1].p-1)->p;
	stkptr+=5;
      } else if(stkptr->l==KW_NEXT) {
	progptr=((CELLTYPE*)stkptr[1].p-1)->p;
	stkptr+=3;
	fstkptr+=2;
      } else doerror(GE_WIT,"EXIT FOR","FOR");
    break;
    /* these are used to push references to on-stack parameters */
    case OP_REFSTK: p=stkptr;ppush(p);break;
    case OP_REFFP:  ppush(fstkptr);break;
    case OP_PARMREF: a=progptr->l;progptr++;p=(frameptr+a)->p;
      ppush(p);break; /* reference parameter access */
    case OP_ILOCREF: a=progptr->l;progptr++;p=frameptr+a;ppush(p);break;
    case OP_SLOCREF: a=progptr->l;progptr++;p=stframe+a;ppush(p);break;
    case OP_FLOCREF: a=progptr->l;progptr++;p=(CELLTYPE*)(flframe+a);
      ppush(p);break;
    case OP_SPUTL2: /* store string function result variable */
     ppop(ca); ca=cat_str(ca,short_strings); /* force it into stack */
     a=progptr->l;progptr++;
     (frameptr+a)->p=ca;strstkmark=strstkptr;break; 
    case OP_IFORL: p=frameptr+progptr->l;goto lab_ifor;
    case OP_FFORL: p=(CELLTYPE*)(flframe+progptr->l);goto lab_ffor;
    case OP_DROP: stkptr++;break;
    case OP_DUP: a=stkptr->l;ipush(a);break;
    case OP_FDUP: fa=*fstkptr;fpush(fa);break;
    case OP_OVER: a=stkptr[1].l;ipush(a);break;
    case OP_2PICK: a=stkptr[2].l;ipush(a);break;
    case OP_INPUTCOM: input_com();break;
    case OP_SINPUTL: ca=linput_str();ppush(ca);break;
    case OP_FOPEN: ppop(ca);ipop(a);ppop(cb); do_fopen(cb,a,ca);break;
    case OP_FCLOSE: ipop(a); do_fclose(a); break;
    case OP_SETINPUT: ipop(a); set_input(a);break;
    case OP_SETOUTPUT: ipop(a); set_output(a);break;
    case OP_EOF: ipop(a);a=do_eof(a);ipush(a);break;
    case OP_LOF: ipop(a);a=do_lof(a);ipush(a);break;
    case OP_FREEFILE: a=do_freefile();ipush(a);break;
    case OP_ROT: t=stkptr[2];stkptr[2]=stkptr[1];stkptr[1]=stkptr[0];
      stkptr[0]=t;break;
    case OP_SWRITE: ppop(ca);write_str(ca);break;
    case OP_MARKLINE: last_line=progptr;progptr+=3;break;
    case OP_SKDATA: progptr=progptr->p;break;
    case OP_STOP: doerror(GE_STP,0,0);break;
    default: doerror(GE_OPC,0,0);
    }
  }
}

/* Array descriptor:
Cell 0: number of dimensions.
Cell 1: pointer to data.
Cell 2:.. dimensions.
*/
void dim_array(int ndim,CELLTYPE **ap,int type)
{
  long a,b,c,d;
  CELLTYPE t;
  a=ndim;
  (*ap)=malloc((ndim+2)*sizeof(CELLTYPE)); /* allocate array descriptor */
  if(!*ap) doerror(GE_ARR,0,0);
  ipop(b);
  (*ap)->l=ndim;
  if(b<0) doerror(GE_SUB,0,0);
  (*ap)[1+a].l=b;b++; 
  while(--a) {
    ipop(c);
    if(c<0) doerror(GE_SUB,0,0);
    (*ap)[1+a].l=c;
    b=b*(c+1);
  }
  if(type==TY_INT) {
    d=b*sizeof(long);
  }else if(type==TY_DOUBLE) {
    d=b*sizeof(double);
  }else if(type==TY_STR) {
    d=b*sizeof(unsigned char *);
  }else d=0;
  (*ap)[1].p=malloc(d);
  if(!(*ap)[1].p)doerror(GE_ARR,0,0);
  for(a=0;a<b;a++) {
    if(type==TY_INT) ((long*)((*ap)[1].p))[a]=0;
    else if(type==TY_DOUBLE) ((double*)((*ap)[1].p))[a]=0;
    else if(type==TY_STR) ((unsigned char **)((*ap)[1].p))[a]=
			    short_strings;
  }
}

void * index_array(int type)
{
  long a,b,c;
  CELLTYPE t;
  CELLTYPE **ap;
  a=progptr->l; /* number of dimensions */
  ap=progptr[1].p; /* pointer to variable */
  if(!*ap) { /* undimensioned array? */
    ipush(10);
    dim_array(1,ap,type); /* dimension array of size 10 on the fly */
  }
  if(a!=(*ap)->l) doerror(GE_SUB,0,0); /* no of dimensions mismatch*/
  ipop(b);
  if(b<0 || b>(*ap)[1+a].l) doerror(GE_SUB,0,0);
  while(--a) {
    b=b*((*ap)[1+a].l+1);
    ipop(c);
    if(c<0 || c>(*ap)[1+a].l) doerror(GE_SUB,0,0);
    b+=c;
  }
  if(type==TY_INT) b*=sizeof(long);
  else if(type==TY_DOUBLE) b*=sizeof(double);
  else if(type==TY_STR) b*=sizeof(unsigned char *);
  return (*ap)[1].p+b;
}

int bas_timer(void)
{
  time_t tt;
  struct tm * tmstruct;
  time(&tt);
  tmstruct=localtime(&tt);
  return 3600*tmstruct->tm_hour+60*tmstruct->tm_min+tmstruct->tm_sec;
}

