/* gplbasic.h
   Copyright (C) 1997 L.C. Benschop, Eindhoven, the Netherlands
   */

#include "strings.h"
#include "gplbasic.h"
#include "errors.h"
#include "token.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

unsigned char * stringareas;

unsigned char short_strings[ 256*3 + 2 ];

static unsigned char * cur_area;


void init_strings(void)
{
  int i;
  unsigned char *p;
  short_strings[0]=0;short_strings[1]=128; /* null string */
  for(i=0,p=short_strings+2;i<256;i++) {
   *p++=1;*p++=128;*p++=i; /* lay out the strings of length 1 */
  } 
}

int comp_str(unsigned char *a,unsigned char *b)
{
 int la,lb,d;
 la=(*a)+256*(*(a+1)&127);
 lb=(*b)+256*(*(b+1)&127);
 d=memcmp(a+2,b+2,la>lb?lb:la);
 if(!d) return la-lb;
 else return d;
}

unsigned char * cat_str(unsigned char *a,unsigned char *b)
{
  unsigned char *c;
  int la,lb,lc;
  la=(*a)+256*(*(a+1)&127);
  lb=(*b)+256*(*(b+1)&127);
  lc=la+lb;
  if( lc > 32767 || string_stack+S_STRINGSTACK - strstkptr < 2 + lc)
    doerror(GE_STS,0,0);
  memcpy(strstkptr+2,a+2,la);
  memcpy(strstkptr+2+la,b+2,lb);
  c=strstkptr;
  *c=lc&255;
  *(c+1)=lc>>8;
  strstkptr+=lc+2;
  return c;
}

unsigned char * left_str(unsigned char * s, int a)
{
 int ls,rl;
 unsigned char *r;
 if(a<0)doerror(GE_ARG,0,0);
 ls=(*s)+256*(*(s+1)&127);
 rl=a;if(rl>ls)rl=ls;
 if( string_stack+S_STRINGSTACK-strstkptr<rl+2) doerror(GE_STS,0,0);
 memcpy(strstkptr+2,s+2,rl);
 r=strstkptr;
 *r=rl&255;
 *(r+1)=rl>>8;
 strstkptr+=rl+2;
 return r;
}

unsigned char * right_str(unsigned char * s, int a)
{
 int ls,rl;
 unsigned char *r;
 if(a<0)doerror(GE_ARG,0,0);
 ls=(*s)+256*(*(s+1)&127);
 rl=a;if(rl>ls)rl=ls;
 if( string_stack+S_STRINGSTACK-strstkptr<rl+2) doerror(GE_STS,0,0);
 memcpy(strstkptr+2,s+2+ls-rl,rl);
 r=strstkptr;
 *r=rl&255;
 *(r+1)=rl>>8;
 strstkptr+=rl+2;
 return r;
}

unsigned char * mid_str2(unsigned char * s, int a)
{
 int ls,rl;
 unsigned char *r;
 if(a<1)doerror(GE_ARG,0,0);
 ls=(*s)+256*(*(s+1)&127);
 rl=ls-a+1;if(rl<0)rl=0;
 if( string_stack+S_STRINGSTACK-strstkptr<rl+2) doerror(GE_STS,0,0);
 memcpy(strstkptr+2,s+2+a-1,rl);
 r=strstkptr;
 *r=rl&255;
 *(r+1)=rl>>8;
 strstkptr+=rl+2;
 return r;
}

unsigned char * mid_str(unsigned char * s, int a,int b)
{
 int ls,rl;
 unsigned char *r;
 if(a<1||b<0)doerror(GE_ARG,0,0);
 ls=(*s)+256*(*(s+1)&127);
 rl=b;if(rl>ls-a+1)rl=ls-a+1;if(rl<0)rl=0;
 if( string_stack+S_STRINGSTACK-strstkptr<rl+2) doerror(GE_STS,0,0);
 memcpy(strstkptr+2,s+2+a-1,rl);
 r=strstkptr;
 *r=rl&255;
 *(r+1)=rl>>8;
 strstkptr+=rl+2;
 return r;
}

unsigned char* str_str(double d)
{
  unsigned char *r;
  int rl;
  if(string_stack+S_STRINGSTACK-strstkptr<32) doerror(GE_STS,0,0);
  sprintf(strstkptr+2,"% .10g ",d);
  rl=strlen(strstkptr+2);
  r=strstkptr;
  *r=rl&255;
  *(r+1)=rl>>8;
  strstkptr+=rl+2;
  return r;
}

unsigned char* oct_str(long l)
{
  unsigned char *r;
  int rl;
  if(string_stack+S_STRINGSTACK-strstkptr<32) doerror(GE_STS,0,0);
  sprintf(strstkptr+2,"%lo",l);
  rl=strlen(strstkptr+2);
  r=strstkptr;
  *r=rl&255;
  *(r+1)=rl>>8;
  strstkptr+=rl+2;
  return r;
}

unsigned char* hex_str(long l)
{
  unsigned char *r;
  int rl;
  if(string_stack+S_STRINGSTACK-strstkptr<32) doerror(GE_STS,0,0);
  sprintf(strstkptr+2,"%lX",l);
  rl=strlen(strstkptr+2);
  r=strstkptr;
  *r=rl&255;
  *(r+1)=rl>>8;
  strstkptr+=rl+2;
  return r;
}

double val_str(unsigned char * s)
{
 double d;
 int ls;
 ls=(*s)+256*(*(s+1)&127);
 if(ls>32) ls=32;
 memcpy(tokname,s+2,ls);
 tokname[ls]=0;
 if(tokname[0]=='&') {
   if(tokname[1]=='h' || tokname[1]=='H' ) {
     d=strtoul(tokname+2,0,16);
   } else if(tokname[1]=='o' || tokname[1]=='O' ) {
     d=strtoul(tokname+2,0,8);
   } else d=0;
 } else d=atof(tokname);
 return d;
}

unsigned char * string_str(int n,int c) 
{
 int i;
 unsigned char * res;
 if( string_stack+S_STRINGSTACK-strstkptr<n+2) doerror(GE_STS,0,0);
 if(n<0 ||n>32767) doerror(GE_ARG,0,0);
 res=strstkptr;
 *strstkptr++=n&255;
 *strstkptr++=n>>8;
 for(i=0;i<n;i++) {
  *strstkptr++=c;
 }
 return res;
}

int instr(int start,unsigned char * haystack, unsigned char * needle) 
{
 int pos,hlen,nlen;
 hlen=*haystack+256*(*(haystack+1)&127);
 nlen=*needle+256*(*(needle+1)&127);
 pos=start-1;if(pos<0)pos=0;
 for(; pos+nlen<=hlen ; pos++ ) {
   if(!memcmp(haystack+pos+2,needle+2,nlen)) return pos+1;
 }
 return 0;
}



/* string area. Start: pointer to next string area.
                Next:  pointer to last free position in area 
	        This is followed by repeated:
	        - back pointer to string variable (sizeof char ** chars).
	        - length of string (2 chars).
	        - characters of string.
	        */

void put_str(unsigned char **var,unsigned char * s)
     /* store string into variable, taking care of space allocation in
        the string heap */
{
  int rl; /*string length */
  int i,rs;
  unsigned char ** bptr,*freptr,*srcptr,*dstptr,*oldcur;
  if(*var==s) return; /* nothing to be done if var already points to s */
  if(!((*var)[1]&128)) { /* original string variable does not point to
                            constant, so back pointer must be destroyed */
    bptr=0;
    memcpy(*var - sizeof(bptr),&bptr,sizeof(bptr)); 
  }
  rl=*s+256*(*(s+1)&127);
  if(rl==0) {
    *var=short_strings;
  } else if (rl==1) {  /* strings of length 0 or 1 are preallocated. */
    *var=short_strings+2+3*s[2];
  } else if(*(s+1)&128) { /* is it a constant string? */
    *var=s;
  } else { /* we must find an allocation for this string */
    if(stringareas==0) { /* nothing was allocated yet? */
      stringareas=malloc(S_STRINGAREA);
      if(stringareas==0) doerror(GE_STR,0,0);
      ((char**)stringareas)[0]=0;
      ((char**)stringareas)[1]=stringareas+2*sizeof(char*);
      cur_area=stringareas;
    }
    freptr=((char**)cur_area)[1];
    if(S_STRINGAREA-rl-2-sizeof(bptr) < freptr-cur_area) { 
               /* no room at the moment in current area. */
      oldcur=cur_area;
      do { /* visit all all string areas in turn to find room for the string */
	cur_area=((char **)cur_area)[0];
	if(cur_area==0) cur_area=stringareas;
	freptr=((char **)cur_area)[1];
      } while(S_STRINGAREA-rl-2-sizeof(bptr) < freptr-cur_area && 
	      cur_area!=oldcur); 
      if(cur_area==oldcur) { /* visited all string areas without success */
	do { /* compact all string areas in turn to make room for the string */
	  cur_area=((char **)cur_area)[0];
	  if(cur_area==0) cur_area=stringareas;
	  freptr=((char **)cur_area)[1];
	  srcptr=cur_area+2*sizeof(char*); /* start garbage collection */
	  dstptr=srcptr;
	  while(srcptr!=freptr) { /* move strings down */
	    memcpy(&bptr,srcptr,sizeof(bptr));
	    rs=srcptr[sizeof(bptr)]+256*srcptr[sizeof(bptr)+1];
	    if (bptr) { /* string is still in use */
	      if(srcptr!=dstptr) { /* move this string down */
		for(i=0;i<sizeof(bptr)+2+rs;i++) {
		  dstptr[i]=srcptr[i];
		}
		if(srcptr+sizeof(bptr)==s)s=dstptr+sizeof(bptr);
		/* if s points to the variable being moved, adjust it */
		*bptr=dstptr+sizeof(bptr);
		/* adjust variable pointed to by back pointer */
	      }
	      dstptr+=sizeof(bptr)+2+rs;
	    }
	    srcptr+=sizeof(bptr)+2+rs;
	  }
	  ((char **)cur_area)[1]=freptr=dstptr;
	} while(S_STRINGAREA-rl-2-sizeof(bptr) < freptr-cur_area &&
		cur_area!=oldcur);
	if(S_STRINGAREA-rl-2-sizeof(bptr) < freptr-cur_area) {
	  /* we have to add a new string area */
	  cur_area=stringareas;
	  stringareas=malloc(S_STRINGAREA);
	  if(stringareas==0) doerror(GE_STR,0,0);
	  ((char**)stringareas)[0]=cur_area;
	  ((char**)stringareas)[1]=freptr=stringareas+2*sizeof(char*);
	  cur_area=stringareas;
	}
      }
    }
    /* at this point we can store the string at freptr */
    bptr=var;
    memcpy(freptr,&bptr,sizeof(bptr)); /* set back pointer */
    freptr+=sizeof(bptr);
    *var=freptr;
    memcpy(freptr,s,rl+2); /* copy string */
    freptr+=rl+2;
    ((char **)cur_area)[1]=freptr;
  }
}

unsigned char * time_str(void)
{
 int rl=8;
 time_t tt;
 struct tm *tmstruct;
 unsigned char *r;
 if( string_stack+S_STRINGSTACK-strstkptr<rl+2+1) doerror(GE_STS,0,0);
 time(&tt);
 tmstruct=localtime(&tt);
 strftime(strstkptr+2,rl+1,"%H:%M:%S",tmstruct);
 r=strstkptr;
 *r=rl&255;
 *(r+1)=rl>>8;
 strstkptr+=rl+2;
 return r;
}

unsigned char * date_str(void)
{
 int rl=10;
 time_t tt;
 struct tm *tmstruct;
 unsigned char *r;
 if( string_stack+S_STRINGSTACK-strstkptr<rl+2+1) doerror(GE_STS,0,0);
 time(&tt);
 tmstruct=localtime(&tt);
 strftime(strstkptr+2,rl+1,"%m-%d-%Y",tmstruct);
 r=strstkptr;
 *r=rl&255;
 *(r+1)=rl>>8;
 strstkptr+=rl+2;
 return r;
}

unsigned char * ucase_str(unsigned char * s)
{
 int rl,i;
 unsigned char c;
 unsigned char *r;
 rl=(*s)+256*(*(s+1)&127);
 if( string_stack+S_STRINGSTACK-strstkptr<rl+2) doerror(GE_STS,0,0);
 for(i=0;i<rl;i++) {
   c=*(s+2+i);
   if(islower(c))c=toupper(c);
   *(strstkptr+2+i)=c;
 }
 r=strstkptr;
 *r=rl&255;
 *(r+1)=rl>>8;
 strstkptr+=rl+2;
 return r;
}

unsigned char * lcase_str(unsigned char * s)
{
 int rl,i;
 unsigned char c;
 unsigned char *r;
 rl=(*s)+256*(*(s+1)&127);
 if( string_stack+S_STRINGSTACK-strstkptr<rl+2) doerror(GE_STS,0,0);
 for(i=0;i<rl;i++) {
   c=*(s+2+i);
   if(isupper(c))c=tolower(c);
   *(strstkptr+2+i)=c;
 }
 r=strstkptr;
 *r=rl&255;
 *(r+1)=rl>>8;
 strstkptr+=rl+2;
 return r;
}

unsigned char * ltrim_str(unsigned char * s)
{
 int rl,sl,i;
 unsigned char *r;
 sl=(*s)+256*(*(s+1)&127);
 for(i=0;i<sl;i++) {
   if(*(s+2+i)!=' ')break;
 }
 rl=sl-i;
 s+=i;
 if( string_stack+S_STRINGSTACK-strstkptr<rl+2) doerror(GE_STS,0,0);
 memcpy(strstkptr+2,s+2,rl);
 r=strstkptr;
 *r=rl&255;
 *(r+1)=rl>>8;
 strstkptr+=rl+2;
 return r;
}

unsigned char * rtrim_str(unsigned char * s)
{
 int rl,sl,i;
 unsigned char *r;
 sl=(*s)+256*(*(s+1)&127);
 for(i=0;i<sl;i++) {
   if(*(s+1+sl-i)!=' ')break;
 }
 rl=sl-i;
 if( string_stack+S_STRINGSTACK-strstkptr<rl+2) doerror(GE_STS,0,0);
 memcpy(strstkptr+2,s+2,rl);
 r=strstkptr;
 *r=rl&255;
 *(r+1)=rl>>8;
 strstkptr+=rl+2;
 return r;
}
