/* token.c
   Copyright (C) 1997 L.C. Benschop, Eindhoven, the Netherlands
   */

#include "gplbasic.h"
#include "token.h"
#include "errors.h"
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

typedef struct kw_struct {
      int num;
      char * name;
      char flags;
      struct kw_struct * next;
} KEYWORD;
/* flag bit 0: occurs in Q(UICK)BASIC.
            1: occurs in BBC BASIC.
	    */

KEYWORD init_kwtab[] = {
  { KW_ABS , "ABS" , 3},
  { KW_ACS , "ACS" , 2},
  { KW_ASN , "ASN" , 2},
  { KW_ATN , "ATN" , 3},
  { KW_COS , "COS" , 3},
  { KW_SIN , "SIN" , 3},
  { KW_TAN , "TAN" , 3},
  { KW_SGN , "SGN" , 3},
  { KW_AND , "AND" , 3},
  { KW_OR  , "OR" , 3},
  { KW_XOR , "XOR" , 1},
  { KW_XOR , "EOR", 2},
  { KW_EQV , "EQV" , 1},
  { KW_IMP , "IMP" , 1},
  { KW_SQR , "SQR" , 3},
  { KW_LOG , "LOG" , 3},
  { KW_LN  , "LN" , 2}, 
  { KW_EXP , "EXP" , 3},
  { KW_RND , "RND" , 3},
  { KW_ASC , "ASC" , 3},
  { KW_CHR , "CHR$" , 3},
  { KW_LEFT , "LEFT$" , 3},
  { KW_RIGHT , "RIGHT$" , 3},
  { KW_MID  , "MID$" , 3},
  { KW_UCASE , "UCASE$", 1},
  { KW_LCASE , "LCASE$", 1},
  { KW_LTRIM , "LTRIM$", 1},
  { KW_RTRIM , "RTRIM$", 1},
  { KW_INSTR , "INSTR" , 3},
  { KW_STRING , "STRING$" , 3},
  { KW_VAL , "VAL" , 3},
  { KW_STR , "STR$" , 3},
  { KW_RUN  , "RUN" , 3},
  { KW_CHAIN , "CHAIN" , 1},
  { KW_GOTO , "GOTO" , 3},
  { KW_GOSUB , "GOSUB" , 3},
  { KW_ON    , "ON" , 3},
  { KW_RETURN , "RETURN" , 3},
  { KW_IF , "IF"  , 3},
  { KW_THEN , "THEN" , 3},
  { KW_ELSE , "ELSE" , 3},
  { KW_ELSEIF , "ELSEIF" , 3},
  { KW_ENDIF , "ENDIF" , 3},
  { KW_DO  , "DO" , 1},
  { KW_EXIT , "EXIT" , 3},
  { KW_LOOP , "LOOP" , 1},
  { KW_WHILE , "WHILE" , 3},
  { KW_WEND , "WEND" , 1},
  { KW_REPEAT , "REPEAT" , 2},
  { KW_UNTIL , "UNTIL" , 3},
  { KW_FOR , "FOR" , 3}, 
  { KW_TO , "TO", 3},
  { KW_STEP , "STEP" , 3},
  { KW_NEXT , "NEXT" , 3},
  { KW_END , "END" , 3},
  { KW_STOP , "STOP" , 3},
  { KW_PRINT , "PRINT" , 3},
  { KW_INPUT , "INPUT" , 3},
  { KW_LET , "LET" , 3},
  { KW_REM , "REM" , 3},
  { KW_READ , "READ" , 3},
  { KW_DATA , "DATA" , 3},
  { KW_RESTORE , "RESTORE" , 3},
  { KW_NOT , "NOT", 3},
  { KW_MOD , "MOD" , 3},
  { KW_LEN , "LEN" , 3},
  { KW_DIM , "DIM" , 3},
  { KW_INT , "INT" , 3},
  { KW_CLEAR , "CLEAR" , 3},
  { KW_RANDOMIZE , "RANDOMIZE" , 3},
  { KW_TAB , "TAB" , 3},
  { KW_DEF , "DEF" , 3},
  { KW_SUB , "SUB" , 1},
  { KW_FUNCTION , "FUNCTION" , 1},
  { KW_DEFINT , "DEFINT" , 1},
  { KW_DEFLNG , "DEFLNG" , 1},
  { KW_DEFSNG , "DEFSNG" , 1},
  { KW_DEFDBL , "DEFDBL" , 1},
  { KW_DEFSTR , "DEFSTR" , 1},
  { KW_SYSTEM , "SYSTEM" , 1},
  { KW_TIME , "TIME$" , 1},
  { KW_DATE , "DATE$" , 1},
  { KW_TIMER , "TIMER" , 1},
  { KW_OCT , "OCT$" , 1},
  { KW_HEX , "HEX$" , 1},
  { KW_CINT , "CINT" , 1},
  { KW_CLNG , "CLNG" , 1},
  { KW_CSNG , "CSNG" , 1},
  { KW_CDBL , "CDBL" , 1},
  { KW_FIX  , "FIX" , 1},
  { KW_SHARED , "SHARED" , 1},
  { KW_STATIC , "STATIC" , 1},
  { KW_DECLARE , "DECLARE" , 1},
  { KW_CALL , "CALL" , 1},
  { KW_OPTION , "OPTION" , 3},
  { KW_BASE , "BASE" ,3},
  { KW_IS , "IS" , 1},
  { KW_AS , "AS" , 1},
  { KW_SELECT , "SELECT" , 1},
  { KW_CASE , "CASE" , 3},
  { KW_OPEN , "OPEN" , 3},
  { KW_CLOSE , "CLOSE" , 3},
  { KW_OUTPUT , "OUTPUT" , 1},
  { KW_APPEND , "APPEND" , 1},
  { KW_LINE , "LINE" , 1}, 
  { KW_EOF , "EOF" , 3},
  { KW_LOF , "LOF" , 1},
  { KW_FREEFILE , "FREEFILE" , 1},
  { KW_WRITE , "WRITE" , 1},
};

KEYWORD * hashed_kwtab[S_HASHENTRIES];

char * parseptr;

int thistoken;

char tokname[MAXLINELENGTH]; 
unsigned int tokhash;    /* hash belonging to TOK_IDENT tokens */         
int vartype;

int nofloat; /* flag to indicate that no float is desired */

void skipblanks(void)
{
  while(*parseptr==' '||*parseptr=='\t')parseptr++;
}

void init_keywords(int f)
     /* f indicates whether to include keywords from q and b dialects */
{
 int i;
 char *p;
 unsigned hash;
 for(i=0; i<sizeof(init_kwtab)/sizeof(KEYWORD);i++) {
   if(init_kwtab[i].flags&f) {
     hash=0;
     for(p=init_kwtab[i].name;*p && *p!='$';p++) {
       hash *= 5;
       hash ^= *p; 
     }
     hash = hash % S_HASHENTRIES; /* compute hash of keyword w/o $*/
     init_kwtab[i].next=hashed_kwtab[hash];
     hashed_kwtab[hash]= &init_kwtab[i]; /* add to hash table */
   }
 }
}


int nexttoken(void)
{
  int i,digs;
  unsigned hash;
  char c;
  KEYWORD *p;
  skipblanks();
  if( isdigit(*parseptr) || (*parseptr=='.' && (!nofloat)) ) {
    i=0;digs=0;
    thistoken=TOK_INT;
    while(isdigit(*parseptr)) {
      tokname[i++]=*parseptr++;
      digs++;
    } /* scan integer part */
    if( !nofloat) {
      if( *parseptr=='.') {
        thistoken=TOK_FLT;
	tokname[i++]=*parseptr++;
	while(isdigit(*parseptr)) {
	  tokname[i++]=*parseptr++;
	  digs++;
	}
      } /* scan fractional part */ 
      if(!digs)doerror(GE_SYN,0,0);
      if(*parseptr=='E' || *parseptr=='e') {
	thistoken=TOK_FLT;
	tokname[i++]=*parseptr++;
	if(*parseptr=='-' || *parseptr=='+' ) {
	  tokname[i++]=*parseptr++;
	}
	while(isdigit(*parseptr)) {
	  tokname[i++]=*parseptr++;
	}
      } /* scan exponential part */
      if(*parseptr=='#'||*parseptr=='!') { 
	/* # or ! suffix indicate double or single */
	parseptr++;thistoken=TOK_FLT;
      }
    }
    tokname[i]=0;
  } else if ( isalpha(*parseptr) ) {
    thistoken=TOK_IDENT;
    hash=0;
    i=0;
    while(isalnum(*parseptr) || *parseptr=='_') {
      c=*parseptr++;
      if(islower(c))c=toupper(c);
      tokname[i++]=c;
      hash*=5;
      hash^=c;
    } /* scan identifier and compute its hash */
    tokhash = hash;
    hash = hash % S_HASHENTRIES;
    vartype=TY_PLAIN;
    if( *parseptr =='%' ) {
      parseptr++;
      vartype=TY_INT;
    } else if( *parseptr =='&' ) {
      parseptr++;
      vartype=TY_LONG;
    } else if( *parseptr =='!' ) {
      parseptr++;
      vartype=TY_SINGLE;
    } else if( *parseptr =='#' ) {
      parseptr++;
      vartype=TY_DOUBLE;
    } else if( *parseptr =='$' ) {
      tokname[i++]='$'; /* add temporary $ to name */
      parseptr++;
      vartype=TY_STR;
    }
    tokname[i]=0;
    p=hashed_kwtab[hash];
    while(p) {
      if(!strcmp(p->name,tokname)) {
        thistoken=p->num;
        break;
      } else p=p->next;
    }
    if(vartype==TY_STR) tokname[i-1]=0; /* remove $ from name */
  } else {
    switch( *parseptr++ ) {
    case '\'': case 0: case '\n': case '\r': thistoken=TOK_EOL; break;
    case '+' : thistoken=TOK_PLUS; break;
    case '-' : thistoken=TOK_MINUS; break;
    case '*' : thistoken=TOK_TIMES; break;
    case '/' : thistoken=TOK_DIV; break;
    case '^' : thistoken=TOK_POW; break;
    case '\\': thistoken=TOK_IDIV; break;
    case '(' : thistoken=TOK_LPAR; break;
    case ')' : thistoken=TOK_RPAR; break;
    case ',' : thistoken=TOK_COMMA; break;
    case ':' : thistoken=TOK_COLON; break;
    case ';' : thistoken=TOK_SEMI; break;
    case '#' : thistoken=TOK_HASH; break;
    case '=' : thistoken=TOK_EQ; break;
    case '<' : 
      if ( *parseptr=='=') {
	parseptr++;thistoken=TOK_LE;
      } else if ( *parseptr=='>') {
	parseptr++;thistoken=TOK_NE;
      } else thistoken=TOK_LT;
    break;
    case '>' : 
      if ( *parseptr=='=') {
	parseptr++;thistoken=TOK_GE;
      } else thistoken=TOK_GT;
    break;
    case '"' : thistoken=TOK_STR; /* parse string literal */
      i=0;
      for(;;) {      
	if( *parseptr == 0 ) {
	  doerror(GE_SYN,0,0);
	} else if( *parseptr == '"') {
	  if( *(parseptr+1) == '"') { /* treat double "" as one */
	    parseptr+=2;tokname[i++]='"';
	  } else {
	    parseptr++;
	    break;
          }
	} else {
	  tokname[i++]=*parseptr++;
	}
      }
      tokname[i]=0;
    break;
    case '&': 
      if(*parseptr=='H'||*parseptr=='h') {
	thistoken=TOK_HEX;
	parseptr++;
	i=0;
	while(isxdigit(*parseptr)) {
	  tokname[i++]=*parseptr++;
	}
	tokname[i]=0;
      } else if(*parseptr=='O' || *parseptr=='o' ) {
	thistoken=TOK_OCT;
	parseptr++;
	i=0;
	while(isdigit(*parseptr)) {
	  tokname[i++]=*parseptr++;
	}
	tokname[i]=0;
      } else thistoken=TOK_UNK;
    break;
    default: thistoken=TOK_UNK; break;
    }
  }
  return thistoken;
}

void require(int t)
     /* require that current token equals t */
{
 if( thistoken != t) doerror(GE_SYN,0,0);
 nexttoken();
}





